/**************************************************************************\
 *  aRT : API R - TerraLib                                                *
 *  Copyright (C) 2003-2005  LEG                                          *
 *                                                                        *
 *  This program is free software; you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation; either version 2 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU Lesser General Public      *
 *  License along with this library.                                      *
\**************************************************************************/

#ifndef ART_LAYER_H_
#define ART_LAYER_H_

#include "aRTcomponent.h"
#include "aRTtheme.h"
#include "aRTdb.h"
#include "SEXPutils.h"
#include "TeDecoderSEXP.h"

#include <TeLayer.h>

#include <sstream>

extern "C" {

using namespace SEXPutils;
		
class aRTlayer : public aRTcomponentLeaf {
	public:
		aRTlayer(TeDatabase* database, string layername);
		~aRTlayer();

		aRTcomponent* Open(SEXP data){ return NULL;}
		aRTcomponent* Open(string theme_name, bool create = false, string viewname = ""){ return NULL; }
		aRTcomponent* Open(string theme_name, bool create, string viewname, SEXP tableNames, SEXP colors);

		void AddElements(string tablename, SEXP data);
		void CreateTable(string Tablename, SEXP attributes, int keyposition);
		void AddTableData(string strTablename, SEXP table, int keyposition);
		
		void Populate(SEXP attributes);
		void InsertNewAttrib(TeAttributeRep, SEXP attrValues);
		void InsertNewAttrib(SEXP attrValues);

		// Add inserts only at the endo of the table. TODO: append
		void AddPoints(SEXP points);
		void AddPolygons(SEXP polset);
		void AddLines(SEXP lineset);
		void AddColumn(SEXP data);
		SEXP GetBox();
		SEXP GetRaster();
		SEXP GetGeometry();
		SEXP GetAttributes();
		SEXP GetData();
		SEXP GetMetric(SEXP options);
		SEXP GetSetMetric(SEXP options);
		SEXP GetRelation(SEXP options, aRTlayer* layer);

		TeGeomRep GeomRep();
		
		void AddRaster(SEXP raster);
		void CreateEmptyTable();

		void AddTable(string Tablename, SEXP table, int keypos);
		SEXP Summary();
		const string Print(); 
		SEXP List() { return R_NilValue; }
		SEXP ShowTables();
		string Class() { return "aRTlayer"; }
		bool IsValid();
		TeLayer* Layer() { return TeUtils::GetLayerFromDatabase(Database, LayerName); }
	private:
		TeDatabase* Database;
		string LayerName;
};

}

#endif
