/**************************************************************************\
 *  aRT : API R - TerraLib                                                *
 *  Copyright (C) 2003-2005  LEG                                          *
 *                                                                        *
 *  This program is free software; you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation; either version 2 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU Lesser General Public      *
 *  License along with this library.                                      *
\**************************************************************************/

/* In this file, we intent to convert some pointers to C++
 * objects and to call their functions. 
 * 
 * All objects in R contains only the pointer to TerraLib data.
 * This code only converts SEXPs to Te* and Te* to SEXPs, 
 * checks if the data is valid, and call the Te functions.
 */

#include "aRTcomponent.h"
#include "aRT.h"
#include "aRTdb.h"
#include "aRTlayer.h"
#include "aRTtheme.h"
#include "SEXPutils.h"
#include "SEXPbasic.h"

extern "C" {
#include <Rdefines.h>
#include <Rinternals.h>

SEXP getObjHandle(SEXP obj)
{
    SEXP sxpHandle = getAttrib( obj, mkString("pointer") );
    if( isNull(sxpHandle) ) error("Object is NULL");
    return sxpHandle;
}

void* getObj(SEXP sxpHandle)
{
    //SEXP sxpHandle = getObjHandle(sxp);
    void* pointer = R_ExternalPtrAddr(sxpHandle);
    if(!pointer) error("External pointer is NULL");
    return pointer;
}

// converts a SEXP SexpValue_ to a Type_* and puts it into Pointer_
#define SEXP_TO_ARTP(SexpValue_, Type_, Pointer_) \
  Pointer_ = (Type_*) getObj(SexpValue_);

#define ARTP_TO_SEXP(Pointer_, SexpValue_)                                     \
  if (Pointer_ == NULL) SexpValue_ = R_NilValue;                               \
  else                                                                         \
  {                                                                            \
    SexpValue_ = R_MakeExternalPtr((void*) Pointer_, mkChar("aRT"),R_NilValue);\
  }

// check if a SEXP Pointer_ is a valid aRTcomponent*.  If not, returns Return_.
// This macro must be inside a function.
#define CHECK_VALID(SPointer_)                              \
{                                                           \
    aRTcomponent* art_component_;                           \
    SEXP_TO_ARTP(SPointer_, aRTcomponent, art_component_);  \
    if( !art_component_ -> IsValid() ) {                    \
        error("Invalid object");                            \
    }                                                       \
}

// declares a function called FuncionName, that receives a SEXP Pointer
// as argument. It converts the Pointer to a Type*, calls the function 
// Call and then prints its returned value.

#define PRINT_FUNCTION_0(FunctionName_, Type_, Call_) \
SEXP FunctionName_(SEXP SPointer)                     \
{                                                     \
    CHECK_VALID(SPointer)                             \
    Type_* obj;                                       \
    SEXP_TO_ARTP(SPointer, Type_, obj);               \
    string result = obj -> Call_;                     \
    Rprintf("%s", result.c_str());                    \
    return R_NilValue;                                \
}

#define VOID_FUNCTION_1(FunctionName_, arg_, Type_, Call_)  \
SEXP FunctionName_(SEXP Pointer, SEXP arg_)                 \
{                                                           \
    CHECK_VALID(Pointer);                                   \
    Type_* obj;                                             \
    SEXP_TO_ARTP(Pointer, Type_, obj);                      \
    obj->Call_;                                             \
    return R_NilValue;                                      \
}

#define SEXP_FUNCTION_0(FunctionName_, Type_, Call_) \
	SEXP FunctionName_(SEXP SPointer) {              \
    CHECK_VALID(SPointer);                           \
    Type_* obj;                                      \
    SEXP_TO_ARTP(SPointer, Type_, obj)               \
    return obj -> Call_;                             \
}

#define SEXP_FUNCTION_1(FunctionName_, arg_, Type_, Call_)  \
SEXP FunctionName_(SEXP Pointer, SEXP arg_)                 \
{                                                           \
    CHECK_VALID(Pointer);                                   \
    Type_* obj;                                             \
    SEXP_TO_ARTP(Pointer, Type_, obj);                      \
    return obj->Call_;                                      \
}

// Declaring some functions using the above macros
	
//              FUNCTION NAME        SEXP ARG     TYPE          CALL
//              ==================== ============ ============= =================================
PRINT_FUNCTION_0(cppPrint,                        aRTcomponent, Print()     )

SEXP_FUNCTION_0(cppList,                          aRTcomponent, List()      ) // ShowDbs, ...
SEXP_FUNCTION_0(cppSummary,                       aRTcomponent, Summary()   )

SEXP_FUNCTION_0(cppShowThemes,                    aRTdb,    ShowThemes()    )
SEXP_FUNCTION_0(cppShowViews,                     aRTdb,    ShowViews()     )
SEXP_FUNCTION_0(cppShowTables,                    aRTlayer, ShowTables()    )
 
SEXP_FUNCTION_0(cppGetBox,                        aRTlayer, GetBox()        )
SEXP_FUNCTION_0(cppGetThemeGeometry,              aRTtheme, GetGeometry()   )
SEXP_FUNCTION_0(cppGetLayerGeometry,              aRTlayer, GetGeometry()   )
SEXP_FUNCTION_0(cppGetThemeData,                  aRTtheme, GetData()       )
SEXP_FUNCTION_0(cppGetLayerData,                  aRTlayer, GetData()       )
SEXP_FUNCTION_0(cppGetThemeAttributes,            aRTtheme, GetAttributes() )
SEXP_FUNCTION_0(cppGetLayerAttributes,            aRTlayer, GetAttributes() )
SEXP_FUNCTION_0(cppGetRaster,                     aRTlayer, GetRaster()     )

VOID_FUNCTION_1(cppCreateEmptyTable, nothing,     aRTlayer, CreateEmptyTable() )

VOID_FUNCTION_1(cppAddPoints,        points,      aRTlayer, AddPoints(points)                   )
VOID_FUNCTION_1(cppAddPolygons,      polygons,    aRTlayer, AddPolygons(polygons)               )
VOID_FUNCTION_1(cppAddLines,         lines,       aRTlayer, AddLines(lines)                     )
VOID_FUNCTION_1(cppSetAR,            restriction, aRTtheme, SetAR(SEXP_TO_STRING(restriction))  )
VOID_FUNCTION_1(cppSetSR,            restriction, aRTtheme, SetSR(SEXP_TO_STRING(restriction))  )
VOID_FUNCTION_1(cppSetTR,            restriction, aRTtheme, SetTR(SEXP_TO_STRING(restriction))  )
VOID_FUNCTION_1(cppCreateViewFromDb, viewname,    aRTdb,    CreateView(SEXP_TO_STRING(viewname)))
VOID_FUNCTION_1(cppAddRaster,        raster,      aRTlayer, AddRaster(raster)                   )
VOID_FUNCTION_1(cppAddColumn,        data,        aRTlayer, AddColumn(data)                     )
VOID_FUNCTION_1(cppAddUser,          userName,    aRT,      AddUser(SEXP_TO_STRING(userName))   )

SEXP_FUNCTION_1(cppGetMetric,        options,     aRTlayer, GetMetric(options)                  )
SEXP_FUNCTION_1(cppGetSetMetric,     options,     aRTlayer, GetSetMetric(options)               )

SEXP cppGetRelation(SEXP SPointer, SEXP SOptions, SEXP SOtherPointer)
{
    CHECK_VALID(SPointer);
	CHECK_VALID(SOtherPointer);

	aRTlayer* obj;
	aRTlayer* otherobj;

	SEXP_TO_ARTP(SPointer, aRTlayer, obj);
	SEXP_TO_ARTP(SOtherPointer, aRTlayer, otherobj);

	return obj->GetRelation(SOptions, otherobj);
}

SEXP cppPP(SEXP value)
{
    printSEXP(value);
	return R_NilValue;
}

// TerraLib delete removes data from database, but here remove 
// TODO: testar cppDeleteTheme, 
VOID_FUNCTION_1( cppDeleteLayer,  layername,   aRTdb,  DeleteLayer(SEXP_TO_STRING(layername)))
VOID_FUNCTION_1( cppDeleteView,   viewname,    aRTdb,  DeleteView(SEXP_TO_STRING(viewname))  )
VOID_FUNCTION_1( cppDeleteTheme,  themename,   aRTdb,  DeleteTheme(SEXP_TO_STRING(themename)))
VOID_FUNCTION_1( cppDeleteDb,     dbname,      aRT,    DeleteDb(SEXP_TO_STRING(dbname))      )

SEXP cppRemove(SEXP SPointer)
{
	CHECK_VALID(SPointer);
	
	aRTcomponent* obj;
	SEXP_TO_ARTP(SPointer, aRTcomponent, obj);
	delete obj;
	return R_NilValue;
}

SEXP cppSilent(SEXP silent)
{
	if(silent != R_NilValue)
	{
		aRTcomponent::Silent = SEXP_TO_INTEGER(silent);
		return silent;
	}

	SEXP sx = allocVector(LGLSXP, 1);
	LOGICAL(sx)[0] = aRTcomponent::Silent;
	return sx;
}

// FIXME: it is not working atually, but it is not being used too :)
SEXP cppConfirm()
{
	char c;
	scanf("%c", &c);

	SEXP ch = allocVector(STRSXP, 1);
	char cc[2];
	cc[0] = c;
	cc[1] = '\0';
    SET_STRING_ELT(ch, 0, mkChar(cc));
	return ch;
}

// create a new aRT object, open a connection with a SGBD and return its pointer
SEXP cppNewaRT(SEXP user, SEXP password, SEXP port, SEXP host)
{
	SEXP pointer;
	// Database authentication parameters
	string strUser       = SEXP_TO_STRING(user);
	string strPasswd     = SEXP_TO_STRING(password);
	port                 = AS_NUMERIC(port); // vector of double
	unsigned int intPort = (unsigned int)NUMERIC_POINTER(port)[0];
	string strHost       = SEXP_TO_STRING(host);

	ARTP_TO_SEXP( new aRT(strUser, strPasswd, intPort, strHost), pointer )
	return pointer;
}

/*
SEXP cppCreateTable(SEXP SPointer, SEXP STablename, SEXP STabledescription, SEXP SKeyposition)
{
	CHECK_VALID(SPointer);

	aRTlayer* obj;

	SEXP_TO_ARTP(SPointer, aRTlayer, obj);
    int intKey = SEXP_TO_INTEGER(SKeyposition);
	string strTablename = SEXP_TO_STRING(STablename);

	obj->CreateTable(strTablename, STabledescription, intKey);
	return R_NilValue;
}*/

SEXP cppAddElements(SEXP SPointer, SEXP STablename, SEXP table)
{
	CHECK_VALID(SPointer);

	aRTlayer* obj;

	SEXP_TO_ARTP(SPointer, aRTlayer, obj);
	string tablename = SEXP_TO_STRING(STablename);
	obj->AddElements(tablename, table);
	return R_NilValue;	
}

/*
SEXP cppAddTable(SEXP SPointer, SEXP tablename, SEXP table, SEXP keyposition)
{
	CHECK_VALID(SPointer);

	aRTlayer* obj;

	SEXP_TO_ARTP(SPointer, aRTlayer, obj)
	//int intKey          = SEXP_TO_INTEGER(keyposition);
	string strTablename = SEXP_TO_STRING(tablename);

//	obj->AddTable(strTablename, table, intKey);
	return R_NilValue;
}*/

// create \in {TRUE,FALSE}
// option is used only to aRTdb to create a layer passing a projection
// all others are ignored
//SEXP_FUNCTION_1(cppOpen,         data,                 aRTcomponent, Open(data) )


SEXP cppOpen(SEXP SPointer, SEXP SData)
{
	CHECK_VALID(SPointer);

	aRTcomponent* obj, *result;
    SEXP pointer;

	SEXP_TO_ARTP(SPointer, aRTcomponent, obj)
	result = obj -> Open(SData);
	ARTP_TO_SEXP(result, pointer);
	return pointer;
}

SEXP cppOpenTheme(SEXP SPointer, SEXP SThemeName, SEXP create, SEXP SViewName, SEXP STables, SEXP colors)
{
    CHECK_VALID(SPointer);

    aRTcomponent* result;
	aRTlayer* obj;
    SEXP pointer;
    string strThemeName = SEXP_TO_STRING(SThemeName);
    string strViewName  = SEXP_TO_STRING(SViewName);

    SEXP_TO_ARTP(SPointer, aRTlayer, obj)
    result = obj -> Open(strThemeName, SEXP_TO_INTEGER(create), strViewName, STables, colors);
    ARTP_TO_SEXP(result, pointer);
    return pointer;
}

SEXP cppOpenThemeFromDb(SEXP SPointer, SEXP SThemeName)
{
    CHECK_VALID(SPointer);

    aRTcomponent* result;
    aRTdb* obj;
    SEXP pointer;
    string strThemeName = SEXP_TO_STRING(SThemeName);

    SEXP_TO_ARTP(SPointer, aRTdb, obj)
    result = obj -> OpenTheme(strThemeName);
    ARTP_TO_SEXP(result, pointer);
    return pointer;
}

/*
SEXP cppDel(SEXP SPointer)
{
// TODO
	CHECK_VALID(SPointer);

	aRTcomponent* obj;
	SEXP_TO_ARTP(SPointer, aRTcomponent, obj)
	cout << obj -> Print() << endl;
	delete(obj);
	return R_NilValue;
}*/

/*
SEXP cppCreateViewFromTheme(SEXP SPointer, SEXP SName)
{
	CHECK_VALID(SPointer);

	aRTtheme* obj;

	SEXP_TO_ARTP(SPointer, aRTtheme, obj);
	aRTdb* dbObj = (aRTdb*)(obj->GetParent());
	string strName = SEXP_TO_STRING(SName);
    dbObj -> CreateView(strName);

    return R_NilValue;
}*/

/*

SEXP cppExistView(SEXP SPointer, SEXP SName)
{
	CHECK_VALID(SPointer);
	aRTdb* obj;
	SEXP_TO_ARTP(SPointer, aRTdb, obj)
	string strName = SEXP_TO_STRING(SName);
	return R_NilValue;
}*/

/*
SEXP cppSaveTheme(SEXP SPointer, SEXP SName)
{
	CHECK_VALID(SPointer);
	aRTtheme* obj;
	string viewName;
	SEXP_TO_ARTP(SPointer, aRTtheme, obj)
	if (!isNull(SName))
		viewName = SEXP_TO_STRING(SName);
	else viewName = "";
	obj -> Save(viewName);
	return R_NilValue;
}*/

/*SEXP cppSelectTables(SEXP SPointer, SEXP SName)
{
	CHECK_VALID(SPointer);
	aRTtheme* obj;
	string tableName;
	SEXP_TO_ARTP(SPointer, aRTtheme, obj)
	obj -> SelectTables(SName);
	return R_NilValue;
}*/

/*
SEXP cppPopulateLayer(SEXP SPointer, SEXP SName)
{
    CHECK_VALID(SPointer);
    aRTlayer* obj;
    SEXP_TO_ARTP(SPointer, aRTlayer, obj)
  
    obj -> Populate(SName);
    return R_NilValue;
}*/

SEXP cppImportShape(SEXP SPointer, SEXP SLayerName, SEXP SFileName, SEXP Sproj)
{
    CHECK_VALID(SPointer);

    aRTdb* objDb;
    aRTcomponent* objComp;
    SEXP_TO_ARTP(SPointer, aRTdb, objDb)
    SEXP pointer;

    string layerName = SEXP_TO_STRING(SLayerName);
    string fileName  = SEXP_TO_STRING(SFileName);

    TeProjection* proj = NULL;
    if (!isNull(Sproj))
	{
        string strDatum = CHAR(STRING_ELT(GET_SLOT(Sproj, install("datum")),0));
        string class_   = CHAR(STRING_ELT(GET_CLASS(Sproj),0));
        TeDatum datum   = TeDatumFactory::make(strDatum);

        if(class_ == "aRTlatlong")
            proj = new TeLatLong(datum);
    }
    else
        proj = new TeNoProjection();

    objComp = objDb -> ImportShape(fileName, layerName, proj);
    ARTP_TO_SEXP(objComp, pointer);

    return pointer;
}

/*
SEXP cppInsertNewAttr(SEXP SPointer, SEXP SName)
{
    //printSEXP(SPointer);
    CHECK_VALID(SPointer);
    aRTlayer* obj;
    SEXP_TO_ARTP(SPointer, aRTlayer, obj)
  
    obj -> InsertNewAttrib(SName);
    return R_NilValue;
}*/

}

